import asyncio
from abc import ABC, abstractmethod
from typing import Unpack

from notte_browser.session import NotteSession
from notte_sdk.types import AgentRunRequestDict

from notte_agent.common.types import AgentResponse


class BaseAgent(ABC):
    def __init__(self, session: NotteSession):
        self.session: NotteSession = session

    @abstractmethod
    async def arun(self, **data: Unpack[AgentRunRequestDict]) -> AgentResponse:
        pass

    def run(self, **data: Unpack[AgentRunRequestDict]) -> AgentResponse:
        return asyncio.run(self.arun(**data))



from __future__ import annotations

import base64
import json
from dataclasses import dataclass
from typing import Any, TypeVar

from litellm import (
    AllMessageValues,
    ChatCompletionAssistantMessage,
    ChatCompletionAssistantToolCall,
    ChatCompletionImageObject,
    ChatCompletionSystemMessage,
    ChatCompletionTextObject,
    ChatCompletionToolMessage,
    ChatCompletionUserMessage,
    ModelResponse,  # type: ignore[reportPrivateImportUsage]
    OpenAIMessageContent,
)
from litellm.utils import token_counter  # type: ignore[reportUnknownVariableType]
from notte_core.common.config import LlmModel, config
from notte_core.common.logging import logger
from notte_core.errors.llm import LLMParsingError
from notte_core.profiling import profiler
from notte_llm.engine import StructuredContent
from pydantic import BaseModel, Field, PrivateAttr
from typing_extensions import override


# Define valid message roles
# /!\ CachedMessage should stay a dataclass. Don't try to make it a BaseModel it will create some weird issues
# Cf error in https://github.com/nottelabs/notte/pull/285.
@dataclass
class CachedMessage:
    """Message with cached token count"""

    message: AllMessageValues
    token_count: int


T = TypeVar("T", bound=BaseModel)


class Conversation(BaseModel):
    """Manages conversation history and message extraction"""

    history: list[CachedMessage] = Field(default_factory=list)
    json_extractor: StructuredContent = Field(default_factory=lambda: StructuredContent(inner_tag="json"))
    autosize: bool = False
    model: str = Field(default_factory=LlmModel.default)
    max_tokens: int | None = config.max_history_tokens
    conservative_factor: float = 0.8

    _total_tokens: int = PrivateAttr(default=0)
    convert_tools_to_assistant: bool = False

    @override
    def model_post_init(self, __context: Any) -> None:
        if self.max_tokens is None:
            self.max_tokens = LlmModel.get_provider(self.model).context_length

    @property
    def default_max_tokens(self) -> int:
        if self.max_tokens is None:
            raise ValueError("max_tokens is not set")
        return self.max_tokens

    @property
    def conservative_max_tokens(self) -> int:
        """Since token count isn't 100% accurate, allow to be
        slightly conservative, to make sure we trim under the total context length"""
        return int(self.default_max_tokens * self.conservative_factor)

    def count_tokens(self, content: AllMessageValues) -> int:
        """Count the number of tokens in a list of messages"""
        return token_counter(model=self.model, messages=[content])

    def total_tokens(self) -> int:
        """Get total tokens in conversation history"""
        return self._total_tokens

    def trim_history_to_fit(self, new_content: AllMessageValues) -> None:
        """Trim history to make room for new content while preserving system messages"""
        if not self.autosize:
            return

        # Always keep system messages
        init_messages: list[CachedMessage] = []
        other_messages: list[CachedMessage] = []
        is_init_msg = True
        for msg in self.history:
            match is_init_msg, msg.message["role"]:
                case True, "system":
                    init_messages.append(msg)
                case True, "user":
                    # keep first user message as init message (need task description)
                    is_init_msg = False
                    init_messages.append(msg)
                case _, _:
                    other_messages.append(msg)

        new_content_tokens = self.count_tokens(new_content)
        init_tokens = sum(msg.token_count for msg in init_messages)
        available_tokens = self.conservative_max_tokens - init_tokens - new_content_tokens

        # Remove oldest non-system messages until we have room
        current_tokens = sum(msg.token_count for msg in other_messages)
        has_trimmed = 0
        while other_messages and current_tokens > available_tokens:
            removed = other_messages.pop(0)
            current_tokens -= removed.token_count
            has_trimmed += 1

        if has_trimmed > 0:
            logger.info(
                f"Trimmed {has_trimmed} message(s) to stay under max token limit (i.e {self.default_max_tokens // 1000}k)"
            )

        self.history = init_messages + other_messages
        self._total_tokens = sum(msg.token_count for msg in self.history)

    def _add_message(self, msg: AllMessageValues) -> None:
        """Internal helper to add a message with token counting"""
        token_count = self.count_tokens(msg)
        if self.autosize:
            self.trim_history_to_fit(msg)
        cached_msg = CachedMessage(message=msg, token_count=token_count)
        self.history.append(cached_msg)
        self._total_tokens += token_count

    def add_system_message(self, content: str) -> None:
        """Add a system message to the conversation"""
        self._add_message(ChatCompletionSystemMessage(role="system", content=content))

    @profiler.profiled()
    def format_image_content(self, image: bytes) -> ChatCompletionImageObject:
        image_str = base64.b64encode(image).decode("utf-8")
        return ChatCompletionImageObject(
            type="image_url",
            image_url={"url": f"data:image/jpeg;base64,{image_str}"},
        )

    @profiler.profiled()
    def format_user_contents(self, contents: list[str | bytes]) -> OpenAIMessageContent:
        return [
            (
                ChatCompletionTextObject(type="text", text=content)
                if isinstance(content, str)
                else self.format_image_content(content)
            )
            for content in contents
        ]

    def add_user_message(self, content: OpenAIMessageContent, image: bytes | None = None) -> None:
        """Add a user message to the conversation"""
        _content: OpenAIMessageContent = content
        if image is not None and isinstance(content, str):
            _content = self.format_user_contents([content, image])
        self._add_message(ChatCompletionUserMessage(role="user", content=_content))

    def add_user_messages(self, contents: list[str | bytes]) -> None:
        """Add a user message to the conversation"""
        _content: OpenAIMessageContent = self.format_user_contents(contents)
        self._add_message(ChatCompletionUserMessage(role="user", content=_content))

    def add_assistant_message(self, content: str) -> None:
        """Add an assistant message to the conversation"""
        self._add_message(ChatCompletionAssistantMessage(role="assistant", content=content))

    def add_tool_message(self, parsed_content: BaseModel, tool_id: str) -> None:
        """Add a tool message to the conversation"""
        content: str = str(parsed_content.model_dump(mode="json", exclude_unset=True))
        if not self.convert_tools_to_assistant:
            self._add_message(
                ChatCompletionToolMessage(
                    role="tool",
                    content=content,
                    tool_call_id=tool_id,
                )
            )
        else:
            # Optional, convert tools to assistant role
            self._add_message(
                ChatCompletionAssistantMessage(
                    role="assistant",
                    content="",
                    tool_calls=[
                        ChatCompletionAssistantToolCall(
                            id=tool_id,
                            type="function",
                            function={
                                "arguments": content,
                                "name": parsed_content.__class__.__name__,
                            },
                        )
                    ],
                )
            )

    def parse_structured_response(self, response: ModelResponse | str, model: type[T]) -> T:
        """Parse a structured response from the LLM into a Pydantic model

        Args:
            response: The LLM model response
            model: The Pydantic model class to parse into

        Returns:
            Instance of the specified Pydantic model

        Raises:
            LLMParsingError: If response cannot be parsed into the model
        """
        if isinstance(response, str):
            return model.model_validate(response)
        if not response.choices:
            raise LLMParsingError("No choices in LLM response")

        choice = response.choices[0]
        # Extract content from either streaming or non-streaming response
        content: str | None = None
        if isinstance(choice, dict):
            message = choice.get("message", {})  # type: ignore[reportUnknownMemberType]
            if isinstance(message, dict):
                content = message.get("content")  # type: ignore[reportUnknownMemberType]
        else:
            content = getattr(choice, "text")

        if not content:
            raise LLMParsingError("No content in LLM response message")

        try:
            if content is None or not isinstance(content, str):
                raise LLMParsingError("No content in LLM response message")
            extracted = self.json_extractor.extract(content)
            return model.model_validate_json(extracted)
        except (json.JSONDecodeError, ValueError) as e:
            raise LLMParsingError(f"Failed to parse response into {model.__name__}: {str(e)}")

    def messages(self) -> list[AllMessageValues]:
        """Get messages in LiteLLM format

        Returns:
            List of messages formatted for LiteLLM

        Note:
            This converts our internal message format to litellm's format.
            litellm only supports 'assistant' role, so we map all roles to that.
        """
        return [msg.message for msg in self.history]

    def reset(self) -> None:
        """Clear all messages from the conversation"""
        self.history.clear()
        self._total_tokens = 0



from typing import Any, Unpack

from notte_core.common.notifier import BaseNotifier
from notte_sdk.types import AgentRunRequestDict
from typing_extensions import override

from notte_agent.common.base import BaseAgent
from notte_agent.common.types import AgentResponse


class NotifierAgent(BaseAgent):
    """Agent wrapper that sends notifications after task completion."""

    def __init__(self, agent: BaseAgent, notifier: BaseNotifier):
        super().__init__(session=agent.session)
        self.agent: BaseAgent = agent
        self.notifier: BaseNotifier = notifier

    def __getattr__(self, name: str) -> Any:
        """Delegate attribute access to the wrapped agent if not found on this instance."""
        return getattr(self.agent, name)

    @override
    async def arun(self, **data: Unpack[AgentRunRequestDict]) -> AgentResponse:
        """Run the agent and send notification about the result."""
        result = await self.agent.arun(**data)
        self.notifier.notify(data["task"], result)  # pyright: ignore [reportArgumentType]
        return result



from notte_core.errors.base import NotteBaseError


class StepExecutionFailure(NotteBaseError):
    def __init__(self, message: str):
        super().__init__(
            user_message=message,
            agent_message=message,
            dev_message=message,
        )


class MaxConsecutiveFailuresError(NotteBaseError):
    def __init__(self, max_failures: int):
        self.max_failures: int = max_failures
        message = f"Max consecutive failures reached in a single step: {max_failures}."
        super().__init__(
            user_message=message,
            agent_message=message,
            dev_message=message,
        )



from typing import final

import chevron
from notte_core.actions import CompletionAction
from notte_core.browser.observation import ExecutionResult, Observation, TrajectoryProgress
from notte_core.trajectory import Trajectory
from notte_llm.engine import LLMEngine
from pydantic import BaseModel, ValidationError

from notte_agent.common.conversation import Conversation
from notte_agent.common.perception import BasePerception

system_rules = """
You are a validator of an agent who interacts with a browser.
Validate if the output of last action is what the user wanted and if the task is completed.
If the task is unclear defined, you can let it pass.
But if something is missing or the image does not show what was requested dont let it pass.
Try to understand the page and help the model with suggestions like scroll, do x, ... to get the solution right.

Task to validate: {{task}}.

Return a JSON object with 2 keys: `is_valid` and `reason`:
- `is_valid` is a boolean that indicates if the output is correct.
- `reason` is a string that explains why it is valid or not.

Example:
```json
{{&example}}
```
"""


class CompletionValidation(BaseModel):
    is_valid: bool
    reason: str


@final
class CompletionValidator:
    def __init__(
        self,
        llm: LLMEngine,
        perception: BasePerception,
        use_vision: bool = True,
        include_attributes: bool = True,
        max_steps: int = 3,
    ):
        self.use_vision = use_vision
        self.include_attributes = include_attributes
        self.llm: LLMEngine = llm
        self.conv: Conversation = Conversation()
        self.perception: BasePerception = perception
        self.max_actions: int = max_steps

    @staticmethod
    def example() -> CompletionValidation:
        return CompletionValidation(
            is_valid=False,
            reason="The user wanted to search for 'cat photos', but the agent searched for 'dog photos' instead.",
        )

    def validation_message(
        self, output: CompletionAction, history: Trajectory, progress: TrajectoryProgress, last_obs: Observation
    ) -> str:
        previous_results = list(history.execution_results())[-self.max_actions :]
        last_actions = "\n".join(self.perception.perceive_action_result(result) for result in previous_results)
        return f"""
I will now provide you some contextual information to help you validate the output of the agent.

Last observation:
{self.perception.perceive(last_obs, progress)}

Last action executions:
{last_actions}

Agent task output to validate:
{output.model_dump_agent()}

Your turn:
"""

    @staticmethod
    def validate_response_format(output: CompletionAction, response_format: type[BaseModel]) -> CompletionValidation:
        """Check that json output fits json schema"""
        try:
            _ = response_format.model_validate_json(output.answer)
        except ValidationError as e:
            return CompletionValidation(
                is_valid=False,
                reason=f"Expecting agent answer to follow the schema {response_format.model_json_schema()}, but found errors: {e.errors()}",
            )
        return CompletionValidation(
            is_valid=True, reason="The output returned by the agent is a valid according to the response format."
        )

    async def validate(
        self,
        task: str,
        output: CompletionAction,
        history: Trajectory,
        progress: TrajectoryProgress,
        response_format: type[BaseModel] | None = None,
    ) -> ExecutionResult:
        """Validate the output of the last action is what the user wanted"""

        # first, validate the output if provided a schema
        if response_format is not None:
            validation = CompletionValidator.validate_response_format(output, response_format)
            if not validation.is_valid:
                return ExecutionResult(action=output, success=False, message=validation.reason)

        observations = list(history.observations())
        if len(observations) == 0:
            return ExecutionResult(action=output, success=False, message="No observations")

        # then, validate that the answer is correct
        last_obs = observations[-1]

        self.conv.reset()
        system_prompt = chevron.render(system_rules, {"task": task, "example": self.example().model_dump_json()})
        self.conv.add_system_message(content=system_prompt)

        validation_message = self.validation_message(output, history, progress, last_obs)

        self.conv.add_user_message(
            content=validation_message,
            image=(last_obs.screenshot.bytes() if self.use_vision else None),
        )

        answer: CompletionValidation = await self.llm.structured_completion(self.conv.messages(), CompletionValidation)
        return ExecutionResult(
            action=output,
            success=answer.is_valid,
            message=answer.reason,
        )



from collections.abc import Awaitable
from types import TracebackType
from typing import Any, Callable, Unpack

from notte_browser.session import NotteSession
from notte_core.actions import BaseAction
from notte_core.actions.typedicts import parse_action
from notte_core.browser.observation import ExecutionResult
from notte_core.common.logging import logger
from notte_core.trajectory import Trajectory
from notte_sdk.types import AgentCreateRequestDict, ExecutionRequestDict

from notte_agent.common.types import AgentResponse
from notte_agent.main import Agent

AGENT_FALLBACK_INSTRUCTIONS = """
goal: {task}
instructions:
- if the goal is unclear or ill-defined, fail immediately and ask the user to clarify the goal.
- only performed the required actions to achieve the goal. Don't take any other action not intended to achieve the goal.
- only a few number of actions should be performed.
- don't navigate to any other page/url except if explicitly asked to do so.
context:
- last action failed with error: {error}
"""


class AgentFallback:
    """
    A context manager that observes a `Session`'s execute calls and triggers an Agent when a step fails.

    Usage:
    ```python
    with notte.AgentFallback(session=session, task="add to cart") as agent:
        session.execute(actions.Click(id="B1"))
        session.execute(actions.Click(id="L3"))
    ```

    The task is the natural language task of the agent (only to be executed if the steps in the context manager fail).
    """

    def __init__(self, session: NotteSession, task: str, **agent_params: Unpack[AgentCreateRequestDict]) -> None:
        self.session: NotteSession = session
        self.trajectory: Trajectory = session.trajectory.view()
        self.task: str = task
        self.steps: list[ExecutionResult] = []
        self.success: bool = True
        self.agent_response: AgentResponse | None = None
        self.agent_params: AgentCreateRequestDict = agent_params

        # Saved originals
        self._orig_aexecute: Callable[..., Awaitable[ExecutionResult]] | None = None
        self._orig_ascrape: Callable[..., Awaitable[Any]] | None = None
        self._agent_invoked: bool = False

    # ------------------------ context manager ------------------------
    def __enter__(self) -> "AgentFallback":
        self._patch_session()
        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc: BaseException | None, tb: TracebackType | None
    ) -> None:
        self._restore_session()
        # If a raw exception escaped user code inside the agent fallback
        if exc is not None and not self._agent_invoked:
            logger.error(f"❌ Unhandled exception in agent fallback: {exc}")
            raise exc
        if self._agent_invoked:
            logger.info(
                f"📚 Agent fallback finished: {self.task} | steps={len(self.steps)} | success={self.success} | agent_invoked={self._agent_invoked}"
            )
        # Do not suppress exceptions if any, but none expected since we capture in wrapper
        return None

    # ------------------------ patching logic ------------------------
    def _patch_session(self) -> None:
        # Save original async execute
        self._orig_aexecute = self.session.aexecute
        self._orig_ascrape = self.session.ascrape

        # scrape is not supported inside the context manager
        async def wrapped_ascrape(*args: Any, **kwargs: Any) -> Any:  # pyright: ignore [reportUnusedParameter]
            raise ValueError(
                "Agent fallback does not support scrape. Please use session.scrape outside of the context manager."
            )

        # Define wrappers
        async def wrapped_aexecute(
            action: BaseAction | dict[str, Any] | None = None,
            *,
            raise_on_failure: bool | None = None,
            **data: Unpack[ExecutionRequestDict],
        ) -> ExecutionResult:
            # Enforce agent fallback constraint
            if raise_on_failure:
                raise ValueError("AgentFallback only supports raise_on_failure=False")

            if isinstance(action, dict):
                action_parsed = parse_action(**action, **data)
            else:
                action_parsed = parse_action(action, **data)

            if self._agent_invoked and self.agent_response is not None:
                logger.warning(f"⚠️ Skipping action: {action_parsed} because agent fallback has been invoked.")
                return ExecutionResult(
                    action=action_parsed,
                    success=True,
                    message="Action skipped because agent fallback has been invoked.",
                    data=None,
                    exception=None,
                )
            # logger.info(f"✏️ Agent fallback executing action: {action_log}")
            # Delegate to original aexecute and do not raise on failure
            result = await self._orig_aexecute(  # type: ignore[misc]
                action=action_parsed, raise_on_failure=False, **data
            )
            # Record and maybe spawn agent
            self._record_step(result)
            if not result.success:
                logger.warning(f"❌ AgentFallback action failed with error: '{result.message}'")
                await self._aspawn_agent_if_needed()
            return result

        # Monkeypatch only aexecute; execute will route through it
        self.session.aexecute = wrapped_aexecute  # pyright: ignore [reportAttributeAccessIssue]
        self.session.ascrape = wrapped_ascrape

    def _restore_session(self) -> None:
        if self._orig_aexecute is not None:
            self.session.aexecute = self._orig_aexecute  # type: ignore[assignment]
        if self._orig_ascrape is not None:
            self.session.ascrape = self._orig_ascrape  # type: ignore[assignment]

    # ------------------------ recording & agent ------------------------
    def _record_step(self, result: ExecutionResult) -> None:
        self.steps.append(result)
        if not result.success:
            self.success = False

    async def _aspawn_agent_if_needed(self) -> None:
        if self._agent_invoked:
            return
        logger.info(f"🤖 Spawning agent fallback after execution failure with task: {self.task}")
        self._agent_invoked = True
        agent = Agent(session=self.session, trajectory=self.trajectory, **self.agent_params)
        self.agent_response = await agent.arun(
            task=AGENT_FALLBACK_INSTRUCTIONS.format(task=self.task, error=self.steps[-1].message)
        )
        if self.agent_response.success:
            logger.info("🔥 Agent succeeded in fixing the execution failure")
            self.success = True
        else:
            logger.error(f"❌ Agent failed to fix the execution failure: {self.agent_response.answer}")



from enum import StrEnum
from typing import Unpack

from notte_browser.session import NotteSession
from notte_browser.tools.base import BaseTool, PersonaTool
from notte_core.common.notifier import BaseNotifier
from notte_core.credentials.base import BaseVault
from notte_core.trajectory import Trajectory
from notte_sdk.endpoints.personas import BasePersona
from notte_sdk.types import AgentCreateRequestDict, AgentRunRequest, AgentRunRequestDict
from typing_extensions import override

from notte_agent.agent import NotteAgent
from notte_agent.common.base import BaseAgent
from notte_agent.common.notifier import NotifierAgent
from notte_agent.common.types import AgentResponse
from notte_agent.falco.agent import FalcoAgent
from notte_agent.gufo.agent import GufoAgent


class AgentType(StrEnum):
    FALCO = "falco"
    GUFO = "gufo"

    def get_agent_class(self) -> type[FalcoAgent | GufoAgent]:
        match self:
            case AgentType.FALCO:
                return FalcoAgent
            case AgentType.GUFO:
                return GufoAgent


class Agent(BaseAgent):
    def __init__(
        self,
        session: NotteSession,
        vault: BaseVault | None = None,
        persona: BasePersona | None = None,
        notifier: BaseNotifier | None = None,
        agent_type: AgentType = AgentType.FALCO,
        trajectory: Trajectory | None = None,
        **data: Unpack[AgentCreateRequestDict],
    ):
        super().__init__(session=session)
        # just validate the request to create type dependency
        self.data: AgentCreateRequestDict = data
        self.vault: BaseVault | None = vault
        self.notifier: BaseNotifier | None = notifier
        self.session: NotteSession = session
        self.agent_type: AgentType = agent_type
        self.trajectory: Trajectory | None = trajectory or session.trajectory.view()

        self.tools: list[BaseTool] = self.session.tools
        if persona is not None:
            self.vault = self.vault or (persona.vault if persona.has_vault else None)
            self.tools.append(PersonaTool(persona))

    def create_agent(
        self,
    ) -> BaseAgent:
        AgentClass = self.agent_type.get_agent_class()
        agent = AgentClass(
            vault=self.vault,
            session=self.session,
            tools=self.tools,
            trajectory=self.trajectory,
            **self.data,
        )

        if self.notifier:
            agent = NotifierAgent(agent, notifier=self.notifier)
        return agent

    @override
    async def arun(self, **data: Unpack[AgentRunRequestDict]) -> AgentResponse:
        agent = self.create_agent()
        # validate args
        res = AgentRunRequest.model_validate(data)
        agent_res = await agent.arun(**res.model_dump())
        if isinstance(agent, NotteAgent):
            self.trajectory = agent.trajectory
        return agent_res



import json
import re
from abc import ABC, abstractmethod
from typing import Literal

from notte_core.actions import (
    BaseAction,
    CompletionAction,
    GotoAction,
    ScrapeAction,
)
from pydantic import BaseModel


class NotteStepAgentOutput(BaseModel):
    observe: GotoAction | None = None
    step: BaseAction | None = None
    scrape: ScrapeAction | None = None
    completion: CompletionAction | None = None

    @property
    def endpoint(self) -> Literal["observe", "step", "scrape", "done"] | None:
        if self.observe is not None:
            return "observe"
        elif self.step is not None:
            return "step"
        elif self.scrape is not None:
            return "scrape"
        elif self.completion is not None:
            return "done"
        else:
            return None

    @property
    def action(self) -> BaseAction | None:
        if self.observe is not None:
            return self.observe
        elif self.step is not None:
            return self.step
        elif self.scrape is not None:
            return self.scrape
        else:
            return None


class ParameterizedAction(BaseModel):
    action_id: str
    params: dict[str, str] | None = None


class BaseParser(ABC):
    @abstractmethod
    def parse(self, text: str) -> NotteStepAgentOutput | None:
        raise NotImplementedError

    @abstractmethod
    def example_format(self, endpoint: Literal["observe", "step", "scrape"]) -> str | None:
        raise NotImplementedError

    @staticmethod
    def search_pattern(text: str, tag: str) -> str | None:
        pattern = re.compile(rf"<{tag}>(.*?)</{tag}>", re.IGNORECASE | re.DOTALL)
        match = pattern.search(text)
        return match.group(1).strip() if match else None

    @staticmethod
    def parse_json(text: str, tag: str | None = None) -> dict[str, str]:
        if tag is not None:
            _text = BaseParser.search_pattern(text, tag)
            if _text is None:
                raise ValueError(f"No text found within <{tag}> tags")
            text = _text
        try:
            data: dict[str, str] = json.loads(text)
        except json.JSONDecodeError:
            raise ValueError("Invalid JSON in action")
        return data



from abc import ABC, abstractmethod


class BasePrompt(ABC):
    @abstractmethod
    def system(self) -> str:
        pass

    @abstractmethod
    def task(self, task: str) -> str:
        pass

    @abstractmethod
    def select_action(self) -> str:
        pass

    @abstractmethod
    def empty_trajectory(self) -> str:
        pass



from notte_core import check_notte_version

from notte_agent.agent_fallback import AgentFallback
from notte_agent.main import Agent

__version__ = check_notte_version("notte_agent")

__all__ = ["Agent", "AgentFallback"]



import json
from pathlib import Path
from typing import Literal

import chevron
from notte_browser.tools.base import BaseTool
from notte_core.actions import (
    BaseAction,
    CaptchaSolveAction,
    ClickAction,
    CompletionAction,
    FormFillAction,
    GotoAction,
    ScrapeAction,
)
from notte_core.errors.processing import InvalidInternalCheckError
from typing_extensions import override

from notte_agent.common.prompt import BasePrompt


class ActionRegistry:
    """Union of all possible actions"""

    def __init__(self, tools: list[BaseTool] | None = None) -> None:
        self.tools: list[BaseTool] = tools or []

    def get_action_map(self) -> dict[str, type[BaseAction]]:
        if len(self.tools) == 0:
            return BaseAction.ACTION_REGISTRY
        actions = {**BaseAction.ACTION_REGISTRY}
        for tool in self.tools:
            actions.update(tool.get_action_map())
        return actions

    def render(self) -> str:
        """Returns a markdown formatted description of all available actions."""
        descriptions: list[str] = []
        actions = self.get_action_map()
        for action_name, action_cls in actions.items():
            try:
                # Get schema and safely remove common fields
                skip_keys = action_cls.non_agent_fields().difference(set(["description"]))
                sub_skip_keys = ["title", "$ref"]
                full_schema = action_cls.model_json_schema()
                schema = {
                    k: {sub_k: sub_v for sub_k, sub_v in v.items() if sub_k not in sub_skip_keys}
                    for k, v in full_schema["properties"].items()
                    if k not in skip_keys
                }
                full_schema["properties"] = schema
                full_schema["additionalProperties"] = False

                if "$defs" in full_schema:
                    del full_schema["$defs"]

                if "description" in full_schema:
                    del full_schema["description"]

                # schema['id'] = schema['id']['default']
                __description: dict[str, str] = schema.pop("description", "No description available")  # type: ignore[type-arg]
                if "default" not in __description:
                    raise InvalidInternalCheckError(
                        check=f"description should have a default value for {action_cls.__name__}",
                        url="unknown url",
                        dev_advice="This should never happen.",
                    )
                _description: str = __description["default"]
                # Format as: ActionName: {param1: {type: str, description: ...}, ...}
                description = f"""
* "{action_name}" : {_description}. Format:
```json
{json.dumps(full_schema)}
```
"""
                descriptions.append(description)
            except Exception as e:
                descriptions.append(f"Error getting schema for {action_cls.__name__}: {str(e)}")

        return "".join(descriptions)


class FalcoPrompt(BasePrompt):
    def __init__(
        self,
        prompt_file: Path | None = None,
        tools: list[BaseTool] | None = None,
    ) -> None:
        if prompt_file is None:
            prompt_file = Path(__file__).parent / "system.md"
        self.system_prompt: str = prompt_file.read_text()
        self.max_actions_per_step: int = 1
        self.tools: list[BaseTool] = tools or []
        self.action_registry: ActionRegistry = ActionRegistry(tools)

    def example_form_filling(self) -> str:
        form_values: dict[Literal["address1", "city", "state"], str] = {
            "address1": "<my address>",
            "city": "<my city>",
            "state": "<my state>",
        }
        return FormFillAction(value=form_values).model_dump_agent_json()  # pyright: ignore [reportArgumentType]

    def example_captcha(self) -> str:
        return CaptchaSolveAction(captcha_type="recaptcha").model_dump_agent_json()

    def example_invalid_sequence(self) -> str:
        return ClickAction(id="X1").model_dump_agent_json()

    def example_navigation_and_extraction(self) -> str:
        return ScrapeAction(
            instructions="Extract the search results from the Google search page"
        ).model_dump_agent_json()

    def completion_example(self) -> str:
        return CompletionAction(success=True, answer="<answer to the task>").model_dump_agent_json()

    def example_step(self) -> str:
        goal_eval = (
            "Analyze the current elements and the image to check if the previous goals/actions"
            " are successful like intended by the task. Ignore the action result. The website is the ground truth. "
            "Also mention if something unexpected happened like new suggestions in an input field. "
            "Shortly state why/why not"
        )
        return chevron.render(
            """
{
  "state": {
    "page_summary": "On the page are company a,b,c wtih their revenue 1,2,3.",
    "relevant_interactions": [{"id": "B2", "reason": "The button with id B2 represents search and I'm looking to search"}],
    "previous_goal_status": "success|failure|unknown",
    "previous_goal_eval": "{{goal_eval}}",
    "memory": "Description of what has been done and what you need to remember until the end of the task",
    "next_goal": "What needs to be done with the next actions"
  },
  "action": {
      "type: "one_action_type",
      // action-specific parameter
      ...
   }
}
""",
            {"goal_eval": goal_eval},
        )

    @override
    def system(self) -> str:
        return (
            chevron.render(
                self.system_prompt,
                {
                    "max_actions_per_step": self.max_actions_per_step,
                    "action_description": self.action_registry.render(),
                    "example_form_filling": self.example_form_filling(),
                    "example_captcha": self.example_captcha(),
                    "example_step": self.example_step(),
                    "completion_example": self.completion_example(),
                    "completion_action_name": CompletionAction.name(),
                    "goto_action_name": GotoAction.name(),
                    "example_navigation_and_extraction": self.example_navigation_and_extraction(),
                    "example_invalid_sequence": self.example_invalid_sequence(),
                },
            )
            + self.system_tools()
        )

    def system_tools(self) -> str:
        if len(self.tools) == 0:
            return ""
        tools_str = "\n".join(tool.instructions() for tool in self.tools)
        return f"""
Additionaly, you have access to the following external tools/information resources:

{tools_str}
"""

    @override
    def task(self, task: str) -> str:
        return f"""
Your ultimate task is: "{task}".
If you achieved your ultimate task, stop everything and use the `completion` action in the next step to complete the task.
If not, continue as usual.
"""

    @override
    def select_action(self) -> str:
        return """Given the previous information, start by reflecting on your last action. Then, summarize the current page and list immediately relevant available interactions (max of 3).
Absolutely do not under any circumstance list or pay attention to any id that is not explicitly found in the page.
From there, select the your next goal, and in turn, your next action.
    """

    @override
    def empty_trajectory(self) -> str:
        return f"""
    No action executed so far...
    Your first action should always be a `{GotoAction.name()}` action with a url related to the task.
    You should reflect what url best fits the task you are trying to solve to start the task, e.g.
    - flight search task => https://www.google.com/travel/flights
    - go to reddit => https://www.reddit.com
    - ...
    ONLY if you have ABSOLUTELY no idea what to do, you can use `https://www.google.com` as the default url.
    THIS SHOULD BE THE LAST RESORT.
    """



from typing import Any

from notte_browser.workflow_variables import Workflow
from notte_core.agent_types import AgentCompletion
from notte_core.common.logging import logger
from notte_sdk.types import AgentRunRequest
from typing_extensions import override

from notte_agent.agent import NotteAgent
from notte_agent.main import Agent


class WorkflowAgent(NotteAgent):
    def __init__(
        self,
        agent: Agent,
        path: str,
    ):
        self.agent: NotteAgent = agent.create_agent()  # pyright: ignore [reportAttributeAccessIssue]
        super().__init__(
            session=self.agent.session,
            prompt=self.agent.prompt,
            perception=self.agent.perception,
            config=self.agent.config,
            vault=self.agent.vault,
            trajectory=self.agent.trajectory,
        )
        with open(path, "r") as f:
            self.workflow: Workflow = Workflow.model_validate_json(f.read())

    @override
    async def observe_and_completion(self, request: AgentRunRequest) -> AgentCompletion:
        current_step = len(self.trajectory)
        nb_workflow_steps = len(self.workflow.steps)

        if current_step == 0 or (
            current_step >= nb_workflow_steps
            or self.trajectory.last_result is None
            or not self.trajectory.last_result.success
        ):
            logger.info("💨 Workflow - using agent mode to complete the trajectory")
            # if the trajectory is complete or the last step failed, reason about the next step
            if current_step == nb_workflow_steps and self.workflow.steps[-1].action.type != "completion":
                request.task = f"{request.task}\n\n CRITICAL: your next action should be a completion action."
            return await super().observe_and_completion(request)
        step = self.workflow.steps[current_step]
        logger.info(f"💨 Workflow - reusing workflow action '{step.action.type}'")
        return AgentCompletion(
            state=step.state,
            action=step.action,
        )

    @override
    async def arun(self, variables: dict[str, Any] | None = None):  # pyright: ignore [reportIncompatibleMethodOverride]
        _ = self.workflow.fill(variables)
        return await super().arun(**self.workflow.request.model_dump())



import datetime as dt
import traceback
import typing

from litellm import AllMessageValues
from notte_browser.session import NotteSession
from notte_browser.vault import VaultSecretsScreenshotMask
from notte_core.actions import (
    BaseAction,
    CaptchaSolveAction,
    CompletionAction,
    FormFillAction,
    GotoAction,
    HelpAction,
)
from notte_core.agent_types import AgentCompletion
from notte_core.browser.observation import ExecutionResult, Observation, TrajectoryProgress
from notte_core.common.config import LlmModel, NotteConfig, RaiseCondition
from notte_core.common.logging import logger
from notte_core.common.telemetry import track_usage
from notte_core.credentials.base import BaseVault, LocatorAttributes
from notte_core.errors.base import ErrorConfig, NotteBaseError
from notte_core.profiling import profiler
from notte_core.trajectory import Trajectory
from notte_llm.engine import LLMEngine
from notte_llm.tracer import LlmUsageDictTracer
from notte_sdk.types import AgentRunRequest, AgentRunRequestDict
from typing_extensions import override

from notte_agent.common.base import BaseAgent
from notte_agent.common.conversation import Conversation
from notte_agent.common.perception import BasePerception
from notte_agent.common.prompt import BasePrompt
from notte_agent.common.types import AgentResponse
from notte_agent.common.validator import CompletionValidator
from notte_agent.errors import MaxConsecutiveFailuresError

# #########################################################
# ############### Possible improvements ###################
# #########################################################

# TODO: improve agent memory (e.g. add a memory manager with RAG)
# TODO: use tooling calling for LLM providers that support it
# TODO: file upload/download
# TODO: DIFF rendering module for DOM changes
# TODO: remove base 64 images from current state (reduce token usage)


class NotteAgent(BaseAgent):
    @track_usage("local.agent.create")
    def __init__(
        self,
        prompt: BasePrompt,
        perception: BasePerception,
        config: NotteConfig,
        session: NotteSession,
        trajectory: Trajectory | None = None,
        vault: BaseVault | None = None,
    ):
        super().__init__(session=session)
        self.config: NotteConfig = config
        self.llm_tracer: LlmUsageDictTracer = LlmUsageDictTracer()
        self.llm: LLMEngine = LLMEngine(model=self.config.reasoning_model, tracer=self.llm_tracer)
        self.perception: BasePerception = perception
        self.prompt: BasePrompt = prompt
        self.trajectory: Trajectory = trajectory or session.trajectory.view()
        self.created_at: dt.datetime = dt.datetime.now()
        self.max_consecutive_failures: int = config.max_consecutive_failures
        self.consecutive_failures: int = 0
        # validator a LLM as a Judge that validates the agent's attempt at completing the task (i.e. `CompletionAction`)
        self.validator: CompletionValidator = CompletionValidator(
            llm=self.llm, perception=self.perception, use_vision=self.config.use_vision
        )
        self.has_run: bool = False

        # ####################################
        # ########### Vault Setup ############
        # ####################################

        # vaults are used to safely input credentials into the sessions without leaking them to the LLM (text + screenshots)
        self.vault: BaseVault | None = vault
        if self.vault is not None:
            # hide vault leaked credentials within llm inputs
            self.llm.structured_completion = self.vault.patch_structured_completion(0, self.vault.get_replacement_map)(  # pyright: ignore [reportAttributeAccessIssue]
                self.llm.structured_completion
            )
            # hide vault leaked credentials within screenshots
            self.session.window.screenshot_mask = VaultSecretsScreenshotMask(vault=self.vault)

    async def action_with_credentials(self, action: BaseAction) -> BaseAction:
        """Replace credentials in the action if the vault contains credentials"""
        if self.vault is not None and self.vault.contains_credentials(action):
            locator = await self.session.locate(action)
            attrs = LocatorAttributes(type=None, autocomplete=None, outerHTML=None)
            if locator is not None:
                # compute locator attributes
                attr_type = await locator.get_attribute("type")
                autocomplete = await locator.get_attribute("autocomplete")
                outer_html = await locator.evaluate("el => el.outerHTML")
                attrs = LocatorAttributes(type=attr_type, autocomplete=autocomplete, outerHTML=outer_html)
                # replace credentials

            if locator is not None or isinstance(action, FormFillAction):
                action = await self.vault.replace_credentials(
                    action,
                    attrs,
                    self.session.snapshot,
                )
        return action

    async def output(self, request: AgentRunRequest, answer: str, success: bool) -> AgentResponse:
        self.trajectory.stop()
        return AgentResponse(
            created_at=self.created_at,
            closed_at=dt.datetime.now(),
            answer=answer,
            success=success,
            trajectory=self.trajectory,
            llm_messages=await self.get_messages(request),
            llm_usage=self.llm_tracer.summary(),
        )

    async def observe_and_completion(self, request: AgentRunRequest) -> AgentCompletion:
        _ = await self.session.aobserve(perception_type=self.perception.perception_type)

        # Get messages with the current observation included
        messages = await self.get_messages(request)

        with ErrorConfig.message_mode("developer"):
            response: AgentCompletion = await self.llm.structured_completion(
                messages,
                response_format=AgentCompletion,
                use_strict_response_format=LlmModel.use_strict_response_format(self.config.reasoning_model),
            )

        await self.trajectory.append(response, force=True)
        return response

    @profiler.profiled()
    @track_usage("local.agent.step")
    async def step(self, request: AgentRunRequest) -> CompletionAction | None:
        """
        Execute a single step of the agent. The flow is as follows:
        1. Observe the current state of the session
        2. Get the messages with the current observation included
        3. Call the LLM with the messages
        4. Append the LLM response to the trajectory
        5. Execute the action
        6. Append the action result to the trajectory
        7. Return the action result if it is a `CompletionAction`
        """
        response = await self.observe_and_completion(request)

        if self.config.verbose:
            logger.trace(f"🔍 LLM response:\n{response}")
        # log the agent state to the terminal
        response.live_log_state()

        # execute the action
        match response.action:
            case HelpAction(reason=reason):
                # Human in the loop is not implemented yet => fail immediately
                error_msg = f"Agent requested human help: {reason}"
                ex_res = ExecutionResult(action=response.action, success=False, message=error_msg)
                await self.trajectory.append(ex_res, force=True)
                return CompletionAction(success=False, answer=error_msg)
            case CaptchaSolveAction(captcha_type=captcha_type) if (
                not self.session.window.resource.options.solve_captchas
            ):
                # if the session doesnt solve captchas => fail immediately
                error_msg = f"Agent encountered {captcha_type} captcha but session doesnt solve captchas: create a session with solve_captchas=True"
                ex_res = ExecutionResult(action=response.action, success=False, message=error_msg)
                await self.trajectory.append(ex_res, force=True)
                return CompletionAction(success=False, answer=error_msg)

            case CompletionAction(success=False, answer=answer):
                # agent decided to stop with failure
                logger.info(f"😭 Agent stopped with failure:\n{answer}")
                result = ExecutionResult(action=response.action, success=False, message=answer)
                await self.trajectory.append(result, force=True)
                return response.action

            case CompletionAction(success=True, answer=answer) as output:
                # need to validate the agent output
                logger.info(f"🔥 Validating agent output:\n{answer}")
                val_result = await self.validator.validate(
                    output=output,
                    history=self.trajectory,
                    task=request.task,
                    progress=self.progress,
                    response_format=request.response_format,
                )
                if val_result.success:
                    # Successfully validated the output
                    logger.info(f"✅ Task completed successfully: {val_result.message}")
                    result = ExecutionResult(action=response.action, success=True, message=val_result.message)
                    await self.trajectory.append(result, force=True)
                    # agent and validator agree, stop with success
                    return response.action

                logger.error(f"🚨 Agent validation failed: {val_result.message}. Continuing...")
                agent_failure_msg = (
                    f"Answer validation failed: {val_result.message}. Agent will continue running. "
                    "CRITICAL: If you think this validation is wrong: argue why the task is finished, or "
                    "perform actions that would prove it is."
                )
                result = ExecutionResult(action=response.action, success=False, message=agent_failure_msg)
                await self.trajectory.append(result, force=True)
                # disagreement between model and validation, continue
                return None
            case _:
                # The action is a regular action => execute it (default case)
                action = await self.action_with_credentials(response.action)
                result = await self.session.aexecute(action, raise_on_failure=False)
                if result.success:
                    self.consecutive_failures = 0
                else:
                    self.consecutive_failures += 1
                    if self.consecutive_failures >= self.max_consecutive_failures:
                        # max consecutive failures reached, raise an exception
                        if result.exception is None:
                            result.exception = ValueError(result.message)
                        raise MaxConsecutiveFailuresError(self.max_consecutive_failures) from result.exception
                # Successfully executed the action => add to trajectory
        step_msg = self.perception.perceive_action_result(result, include_ids=True)
        logger.info(f"{step_msg}\n\n")

        return None

    @property
    def progress(self) -> TrajectoryProgress:
        return TrajectoryProgress(current_step=self.trajectory.num_steps, max_steps=self.config.max_steps)

    @track_usage("local.agent.messages.get")
    @profiler.profiled()
    async def get_messages(self, request: AgentRunRequest) -> list[AllMessageValues]:
        """
        Formats a trajectory into a list of messages for the LLM, including the current observation.

        For every resonning model step, the conversation is reset.
        The conversation follows the following format:

        ### Setup messages
        - [system]        : system prompt containing the initial instructions + action tool calls info
        - [user]          : user request task (e.g. "Find the latest news about AI")
        ### Trajectory messages (one for every step in the trajectory)
            - [assistant] : agent step response (containing memory,state & next action to take)
            - [user]      : session step execution result (success/failure + info message)
        ### Current DOM perception & final intent message
        - [user]          : current DOM perception (browser metadata, page DOM elements, interactive actions, etc.)
        - [user]          : final intent message (e.g. "Select the best action based on whatever I gave you in before")

        /!\\ If `use_vision` is enabled, the DOM perception message will contain a screenshot of the page.
        """
        conv: Conversation = Conversation(
            convert_tools_to_assistant=True, autosize=True, model=self.config.reasoning_model
        )

        if request.url is not None and self.session.page.url != request.url:
            request.task = f"Go to '{request.url}' and {request.task}"

        if request.response_format is not None:
            request.task = f"{request.task}. \n Use the following response format to format your answer:\n```json\n{request.response_format.model_json_schema()}\n```"

        system_msg, task_msg = self.prompt.system(), self.prompt.task(request.task)
        # max steps instructions
        max_steps_msg = (
            f"You have {self.config.max_steps} steps to complete the task. "
            "If you reach the max steps limit and you haven't completed the task, stop immediately and error out.\n"
        )
        task_msg = f"{task_msg} {max_steps_msg}"
        # tool instructions
        if self.vault is not None:
            system_msg += "\n" + self.vault.instructions()
        # storage instructions
        if self.session.storage is not None:
            # Instructions contain the list of uploaded files -> append to the user message
            task_msg = f"{task_msg} {await self.session.storage.instructions()}"

        conv.add_system_message(content=system_msg)
        conv.add_user_message(content=task_msg)

        # otherwise, add all past trajectorysteps to the conversation
        for step in self.trajectory:
            match step:
                case AgentCompletion():
                    # TODO: choose if we want this to be an assistant message or a tool message
                    # self.conv.add_tool_message(step.agent_response, tool_id="step")
                    conv.add_assistant_message(
                        step.model_dump_json(exclude_none=True, context=dict(hide_interactions=True))
                    )
                case ExecutionResult():
                    # add step execution status to the conversation
                    conv.add_user_message(
                        content=self.perception.perceive_action_result(step, include_ids=False, include_data=True)
                    )

                # observation or screenshot
                case _:
                    # TODO: add partial info for previous?
                    pass

        # Add current observation (only if it's not empty)
        last_obs = self.trajectory.last_observation
        if last_obs is not None and last_obs is not Observation.empty():
            perceived_content = self.perception.perceive(obs=last_obs, progress=self.progress)
            image = last_obs.screenshot.bytes() if self.config.use_vision else None
            conv.add_user_message(
                content=perceived_content,
                image=image,
            )
            conv.add_user_message("<WEBSITE_CONTENT_END>")
            conv.add_user_message(self.prompt.select_action())

        # if no action execution in trajectory, add the start trajectory message
        last_exec = self.trajectory.last_result
        if last_exec is None:
            conv.add_user_message(content=self.prompt.empty_trajectory())

        return conv.messages()

    @profiler.profiled()
    @track_usage("local.agent.run")
    @override
    async def arun(self, **data: typing.Unpack[AgentRunRequestDict]) -> AgentResponse:
        if self.has_run:
            raise ValueError("Agent can only be ran once, recreate a new agent to run a second time")

        self.has_run = True
        request = AgentRunRequest.model_validate(data)
        logger.trace(f"Running task: {request.task}")
        self.consecutive_failures = 0
        self.created_at = dt.datetime.now()
        try:
            with ErrorConfig.message_mode("agent"):
                return await self._run(request)
        except NotteBaseError as e:
            if self.config.raise_condition is RaiseCondition.NEVER:
                return await self.output(
                    request, f"Failed due to notte base error: {e.dev_message}:\n{traceback.format_exc()}", False
                )
            logger.error(f"Error during agent run: {e.dev_message}")
            raise e
        except Exception as e:
            if self.config.raise_condition is RaiseCondition.NEVER:
                return await self.output(request, f"Failed due to {e}: {traceback.format_exc()}", False)
            raise e
        finally:
            # in case we failed in step, stop it (relevant for session)
            _ = await self.trajectory.stop_step(ignore_not_in_step=True)
            _ = self.trajectory.stop()

    async def _run(self, request: AgentRunRequest) -> AgentResponse:
        """Execute the task with maximum number of steps"""
        # initial goto, don't do an llm call just for accessing the first page
        if request.url is not None:
            _ = await self.trajectory.start_step()
            _ = await self.session.aobserve(perception_type=self.perception.perception_type)
            await self.trajectory.append(AgentCompletion.initial(request.url), force=True)
            _ = await self.session.aexecute(GotoAction(url=request.url))
            _ = await self.trajectory.stop_step()

        step = 0
        while self.trajectory.num_steps < self.config.max_steps:
            step += 1
            logger.info(f"💡 Step {step}")

            _ = await self.trajectory.start_step()
            completion_action = await self.step(request)
            _ = await self.trajectory.stop_step()

            if completion_action is not None:
                return await self.output(request, completion_action.answer, completion_action.success)

        error_msg = f"Failed to solve task in {self.config.max_steps} steps"
        logger.info(f"🚨 {error_msg}")
        return await self.output(request, error_msg, False)



from notte_core.common.config import PerceptionType
from notte_core.space import ActionSpace
from typing_extensions import override

from notte_agent.falco.perception import FalcoPerception


class GufoPerception(FalcoPerception):
    @property
    @override
    def perception_type(self) -> PerceptionType:
        return "deep"

    @override
    def perceive_actions(self, space: ActionSpace) -> str:
        # same as falco, but use markdown description instead of html
        return space.markdown



from typing import ClassVar

from notte_core.actions import InteractionAction
from notte_core.browser.observation import ExecutionResult, Observation, TrajectoryProgress
from notte_core.browser.snapshot import SnapshotMetadata
from notte_core.common.config import PerceptionType
from notte_core.data.space import DataSpace
from notte_core.space import ActionSpace
from typing_extensions import override

from notte_agent.common.perception import BasePerception, trim_message


class FalcoPerception(BasePerception):
    DISCLAIMER: ClassVar[str] = (
        """You will see the following only once. If you need to remember it and you dont know it yet, write it down in the memory."""
    )

    def __init__(self, with_disclaimer: bool = True):
        self.with_disclaimer: bool = with_disclaimer

    @property
    @override
    def perception_type(self) -> PerceptionType:
        return "fast"

    @override
    def perceive_metadata(self, metadata: SnapshotMetadata, progress: TrajectoryProgress) -> str:
        return f"""
* Current url: {metadata.url}
* Current page title: {metadata.title}
* Current date and time: {metadata.timestamp.strftime("%Y-%m-%d %H:%M:%S")}
* Available tabs:
{metadata.tabs}
* Current step: {progress.current_step}/{progress.max_steps}
"""

    @override
    def perceive(self, obs: Observation, progress: TrajectoryProgress) -> str:
        px_above = obs.metadata.viewport.pixels_above
        px_below = obs.metadata.viewport.pixels_below

        more_above = f"... {px_above} pixels above - scroll or scrape content to see more ..."
        more_below = f"... {px_below} pixels below - scroll or scrape content to see more ..."
        return f"""
{self.DISCLAIMER if self.with_disclaimer else ""}

<WEBSITE_CONTENT_BEGIN>
[Relevant metadata]
{self.perceive_metadata(obs.metadata, progress)}

[Interaction elements and context]
[Start of page]
{more_above if px_above > 0 else ""}
{self.perceive_actions(obs.space)}
{more_below if px_below > 0 else ""}
[End of page]
"""

    @override
    def perceive_actions(self, space: ActionSpace) -> str:
        return space.description

    @override
    def perceive_data(self, data: DataSpace | None, only_structured: bool = True) -> str:
        if data is None:
            return ""
        if only_structured:
            structured_data = data.structured
            if structured_data is None or not structured_data.success or structured_data.data is None:
                error_msg = f" with error: {structured_data.error}" if structured_data is not None else ""
                return f"Scraping failed{error_msg}. Please try again with different instructions."
            return f"""
Extracted JSON data:
```json
{structured_data.data.model_dump_json()}
```
"""
        return f"""
Data scraped from current page view:
```markdown
{data.markdown or "No valid data to display"}
```
"""

    @override
    def perceive_action_result(
        self,
        result: ExecutionResult,
        include_ids: bool = False,
        include_data: bool = True,
    ) -> str:
        id_str = f" with id={result.action.id}" if include_ids and isinstance(result.action, InteractionAction) else ""
        if not result.success:
            err_msg = trim_message(result.message)
            return f"❌ action '{result.action.name()}'{id_str} failed with error: {err_msg}"
        success_msg = f"✅ action '{result.action.name()}'{id_str} succeeded: '{result.action.execution_message()}'"
        if include_data:
            return f"{success_msg}{self.perceive_data(result.data, only_structured=True)}"
        return success_msg



from abc import ABC, abstractmethod

from notte_core.browser.observation import ExecutionResult, Observation, TrajectoryProgress
from notte_core.browser.snapshot import SnapshotMetadata
from notte_core.common.config import PerceptionType, config
from notte_core.common.logging import logger
from notte_core.data.space import DataSpace
from notte_core.space import ActionSpace


def trim_message(message: str, max_length: int | None = config.max_error_length) -> str:
    if max_length is None or len(message) <= max_length:
        return message
    logger.warning(f"Trimming message ({message[:100]})")
    return f"...{message[-max_length:]}"


class BasePerception(ABC):
    @property
    @abstractmethod
    def perception_type(self) -> PerceptionType:
        pass

    @abstractmethod
    def perceive_metadata(self, metadata: SnapshotMetadata, progress: TrajectoryProgress) -> str:
        pass

    @abstractmethod
    def perceive_actions(self, space: ActionSpace) -> str:
        pass

    @abstractmethod
    def perceive_data(self, data: DataSpace | None, only_structured: bool = True) -> str:
        pass

    @abstractmethod
    def perceive(self, obs: Observation, progress: TrajectoryProgress) -> str:
        pass

    @abstractmethod
    def perceive_action_result(
        self,
        result: ExecutionResult,
        include_ids: bool = False,
        include_data: bool = True,
    ) -> str:
        pass



import datetime as dt
import json
from pathlib import Path
from typing import Annotated, Any, Literal

from litellm import AllMessageValues
from notte_core.agent_types import AgentCompletion
from notte_core.browser.observation import Screenshot
from notte_core.common.config import ScreenshotType, config
from notte_core.common.logging import logger
from notte_core.trajectory import Trajectory
from notte_core.utils.webp_replay import ScreenshotReplay, WebpReplay
from notte_llm.tracer import LlmUsageDictTracer
from pydantic import BaseModel, Field, computed_field
from typing_extensions import override


class AgentResponse(BaseModel):
    class Config:
        arbitrary_types_allowed: bool = True

    success: bool
    answer: str
    trajectory: Trajectory = Field(exclude=True)
    # logging information
    created_at: Annotated[dt.datetime, Field(description="The creation time of the agent")]
    closed_at: Annotated[dt.datetime, Field(description="The closed time of the agent")]
    status: str = "closed"
    # only used for debugging purposes
    llm_messages: list[AllMessageValues]
    llm_usage: LlmUsageDictTracer.AggregatedUsage | None

    @computed_field
    @property
    def duration_in_s(self) -> float:
        return (self.closed_at - self.created_at).total_seconds()

    @computed_field
    @property
    def steps(self) -> list[AgentCompletion]:
        return list(self.trajectory.agent_completions())

    @override
    def __str__(self) -> str:
        return (
            f"AgentResponse(success={self.success}, duration_in_s={round(self.duration_in_s, 2)}, answer={self.answer})"
        )

    def screenshots(self) -> list[Screenshot]:
        return [obs.screenshot for obs in self.trajectory.observations()]

    def replay(self, step_texts: bool = True, screenshot_type: ScreenshotType = config.screenshot_type) -> WebpReplay:
        screenshots: list[bytes] = []
        texts: list[str] = []

        for bundle in self.trajectory.step_iterator():
            if bundle.observation is None or bundle.agent_completion is None:
                continue

            screenshots.append(bundle.observation.screenshot.bytes(screenshot_type))
            texts.append(bundle.agent_completion.state.next_goal)

        if len(screenshots) == 0:
            raise ValueError("No screenshots found in agent trajectory")

        if step_texts:
            return ScreenshotReplay.from_bytes(screenshots).get(step_text=texts)  # pyright: ignore [reportArgumentType]
        else:
            return ScreenshotReplay.from_bytes(screenshots).get()

    def save_actions(self, file_path: str, id_type: Literal["selector", "id"] = "selector") -> None:
        if not file_path.endswith(".json"):
            raise ValueError("File path must end with .json")
        path = Path(file_path)
        path.parent.mkdir(parents=True, exist_ok=True)
        actions: list[dict[str, Any]] = []
        for step in self.trajectory:
            match step:
                case AgentCompletion(action=action):
                    exclude_fields = action.non_agent_fields()
                    if id_type == "selector":
                        exclude_fields.remove("selector")
                        exclude_fields.remove("option_selector")
                    exclude_fields.remove("text_label")
                    if "value" in exclude_fields:
                        exclude_fields.remove("value")
                    logger.info(f"Excluding fields: {exclude_fields}")
                    actions.append(step.action.model_dump(exclude=exclude_fields, exclude_none=True))
                case _:
                    # skip other steps
                    pass
            # actions.append(result.action)
        with open(path, "w") as f:
            json.dump(dict(actions=actions), f, indent=4)

    @override
    def __repr__(self) -> str:
        return self.__str__()



from pathlib import Path

from notte_browser.tools.base import BaseTool

from notte_agent.falco.prompt import FalcoPrompt


class GufoPrompt(FalcoPrompt):
    def __init__(self, tools: list[BaseTool] | None = None):
        super().__init__(prompt_file=Path(__file__).parent / "system.md", tools=tools)



import typing

from notte_browser.session import NotteSession
from notte_browser.tools.base import BaseTool
from notte_core.common.config import NotteConfig
from notte_core.credentials.base import BaseVault
from notte_core.trajectory import Trajectory
from notte_sdk.types import AgentCreateRequest, AgentCreateRequestDict

from notte_agent.agent import NotteAgent
from notte_agent.gufo.perception import GufoPerception
from notte_agent.gufo.prompt import GufoPrompt


class GufoAgent(NotteAgent):
    def __init__(
        self,
        session: NotteSession,
        vault: BaseVault | None = None,
        tools: list[BaseTool] | None = None,
        trajectory: Trajectory | None = None,
        **data: typing.Unpack[AgentCreateRequestDict],
    ):
        _ = AgentCreateRequest.model_validate(data)
        config: NotteConfig = NotteConfig.from_toml(**data)
        super().__init__(
            prompt=GufoPrompt(tools=tools),
            perception=GufoPerception(),
            config=config,
            session=session,
            trajectory=trajectory,
            vault=vault,
        )
